--------------------------------------------------------------------------------
-- Regression tests for the enumerant spec conversion
--------------------------------------------------------------------------------

testSpecString :: String
testSpecString = unlines [
   "typemask mask:",
   "\tMASK_A",
   "\tMASK_B",
   "\tMASK_C",
   "\tMASK_D",
   "\tMASK_E",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum LEFT",
   "\tRIGHT",
   "",
   "3DFX_TypeEnum enum:",
   "\tENUM_A",
   "\tENUM_B",
   "\tENUM_C = 42",
   "\tENUM_D",
   "\tuse SGI_foo_bar DEFINE_C",
   "\tENUM_F",
   "\tuse typemask MASK_D",
   "\tENUM_H",
   "\tENUM_I = $MASK_E",
   "\tENUM_J = $ENUM_F + 42",
   "\tLEFT",
   "\tuse typemask RIGHT",
   "",
   "SGI_foo_bar define:",
   "\tDEFINE_A",
   "\tDEFINE_B",
   "\tDEFINE_C = 128",
   "\tDEFINE_D",
   "\tDEFINE_E",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum ENUM_H"
 ]

testExpandedString :: String
testExpandedString = unlines [
   "typemask mask:",
   "\tMASK_A = 1",
   "\tMASK_B = 2",
   "\tMASK_C = 4",
   "\tMASK_D = 8",
   "\tMASK_E = 16",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum LEFT",
   "\tRIGHT = 32",
   "",
   "3DFX_TypeEnum enum:",
   "\tENUM_A = 0",
   "\tENUM_B = 1",
   "\tENUM_C = 42",
   "\tENUM_D = 2",
   "\tuse SGI_foo_bar DEFINE_C",
   "\tENUM_F = 3",
   "\tuse typemask MASK_D",
   "\tENUM_H = 4",
   "\tENUM_I = $MASK_E",
   "\tENUM_J = $ENUM_F + 42",
   "\tLEFT = 5",
   "\tuse typemask RIGHT",
   "",
   "SGI_foo_bar define:",
   "\tDEFINE_A = 0",
   "\tDEFINE_B = 0",
   "\tDEFINE_C = 128",
   "\tDEFINE_D = 0",
   "\tDEFINE_E = 0",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum ENUM_H"
 ]

testEvaluatedString :: String
testEvaluatedString = unlines [
   "typemask mask:",
   "\tMASK_A = 1",
   "\tMASK_B = 2",
   "\tMASK_C = 4",
   "\tMASK_D = 8",
   "\tMASK_E = 16",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum LEFT",
   "\tRIGHT = 32",
   "",
   "3DFX_TypeEnum enum:",
   "\tENUM_A = 0",
   "\tENUM_B = 1",
   "\tENUM_C = 42",
   "\tENUM_D = 2",
   "\tuse SGI_foo_bar DEFINE_C",
   "\tENUM_F = 3",
   "\tuse typemask MASK_D",
   "\tENUM_H = 4",
   "\tENUM_I = 16",
   "\tENUM_J = 45",
   "\tLEFT = 5",
   "\tuse typemask RIGHT",
   "",
   "SGI_foo_bar define:",
   "\tDEFINE_A = 0",
   "\tDEFINE_B = 0",
   "\tDEFINE_C = 128",
   "\tDEFINE_D = 0",
   "\tDEFINE_E = 0",
   "\tuse 3DFX_TypeEnum ENUM_C",
   "\tuse 3DFX_TypeEnum ENUM_H"
 ]

testSimplifiedString :: String
testSimplifiedString = unlines [
   "Typemask mask:",
   "\tMaskA = 1",
   "\tMaskB = 2",
   "\tMaskC = 4",
   "\tMaskD = 8",
   "\tMaskE = 16",
   "\tEnumC = 42",
   "\tLeft = 5",
   "\tRight = 32",
   "",
   "ThreeDFXTypeEnum enum:",
   "\tEnumA = 0",
   "\tEnumB = 1",
   "\tEnumC = 42",
   "\tEnumD = 2",
   "\tDefineC = 128",
   "\tEnumF = 3",
   "\tMaskD = 8",
   "\tEnumH = 4",
   "\tEnumI = 16",
   "\tEnumJ = 45",
   "\tLeft = 5",
   "\tRight = 32",
   "",
   "SGIFooBar define:",
   "\tdefineA = 0",
   "\tdefineB = 0",
   "\tdefineC = 128",
   "\tdefineD = 0",
   "\tdefineE = 0",
   "\tenumC = 42",
   "\tenumH = 4"
 ]

testRenamedString :: String
testRenamedString = unlines [
   "Typemask mask:",
   "\tMaskA = 1",
   "\tMaskB = 2",
   "\tMaskC = 4",
   "\tMaskD = 8",
   "\tMaskE = 16",
   "\tEnumC' = 42",
   "\tLeft'2 = 5",
   "\tRight' = 32",
   "",
   "ThreeDFXTypeEnum enum:",
   "\tEnumA = 0",
   "\tEnumB = 1",
   "\tEnumC = 42",
   "\tEnumD = 2",
   "\tDefineC = 128",
   "\tEnumF = 3",
   "\tMaskD' = 8",
   "\tEnumH = 4",
   "\tEnumI = 16",
   "\tEnumJ = 45",
   "\tLeft' = 5",
   "\tRight'2 = 32",
   "",
   "SGIFooBar define:",
   "\tdefineA = 0",
   "\tdefineB = 0",
   "\tdefineC = 128",
   "\tdefineD = 0",
   "\tdefineE = 0",
   "\tenumC = 42",
   "\tenumH = 4"
 ]

testCodeString :: String
testCodeString = show $ unlines [
   "--------------------------------------------------",
   "data Typemask =",
   "     MaskA",
   "   | MaskB",
   "   | MaskC",
   "   | MaskD",
   "   | MaskE",
   "   | EnumC'",
   "   | Left'2",
   "   | Right'",
   "",
   "marshalTypemask :: Typemask -> GLbitfield",
   "marshalTypemask x = case x of",
   "   MaskA -> 0x1",
   "   MaskB -> 0x2",
   "   MaskC -> 0x4",
   "   MaskD -> 0x8",
   "   MaskE -> 0x10",
   "   EnumC' -> 0x2a",
   "   Left'2 -> 0x5",
   "   Right' -> 0x20",
   "",
   "unmarshalTypemask :: GLbitfield -> Typemask",
   "unmarshalTypemask x",
   "   | x == 0x1 = MaskA",
   "   | x == 0x2 = MaskB",
   "   | x == 0x4 = MaskC",
   "   | x == 0x8 = MaskD",
   "   | x == 0x10 = MaskE",
   "   | x == 0x2a = EnumC'",
   "   | x == 0x5 = Left'2",
   "   | x == 0x20 = Right'",
   "   | otherwise = error (\"unmarshalTypemask: illegal value \" ++ show x)",
   "",
   "--------------------------------------------------",
   "data ThreeDFXTypeEnum =",
   "     EnumA",
   "   | EnumB",
   "   | EnumC",
   "   | EnumD",
   "   | DefineC",
   "   | EnumF",
   "   | MaskD'",
   "   | EnumH",
   "   | EnumI",
   "   | EnumJ",
   "   | Left'",
   "   | Right'2",
   "",
   "marshalThreeDFXTypeEnum :: ThreeDFXTypeEnum -> GLenum",
   "marshalThreeDFXTypeEnum x = case x of",
   "   EnumA -> 0x0",
   "   EnumB -> 0x1",
   "   EnumC -> 0x2a",
   "   EnumD -> 0x2",
   "   DefineC -> 0x80",
   "   EnumF -> 0x3",
   "   MaskD' -> 0x8",
   "   EnumH -> 0x4",
   "   EnumI -> 0x10",
   "   EnumJ -> 0x2d",
   "   Left' -> 0x5",
   "   Right'2 -> 0x20",
   "",
   "unmarshalThreeDFXTypeEnum :: GLenum -> ThreeDFXTypeEnum",
   "unmarshalThreeDFXTypeEnum x",
   "   | x == 0x0 = EnumA",
   "   | x == 0x1 = EnumB",
   "   | x == 0x2a = EnumC",
   "   | x == 0x2 = EnumD",
   "   | x == 0x80 = DefineC",
   "   | x == 0x3 = EnumF",
   "   | x == 0x8 = MaskD'",
   "   | x == 0x4 = EnumH",
   "   | x == 0x10 = EnumI",
   "   | x == 0x2d = EnumJ",
   "   | x == 0x5 = Left'",
   "   | x == 0x20 = Right'2",
   "   | otherwise = error (\"unmarshalThreeDFXTypeEnum: illegal value \" ++ show x)",
   "",
   "--------------------------------------------------",
   "-- SGIFooBar",
   "defineA :: GLenum",
   "defineA = 0",
   "defineB :: GLenum",
   "defineB = 0",
   "defineC :: GLenum",
   "defineC = 128",
   "defineD :: GLenum",
   "defineD = 0",
   "defineE :: GLenum",
   "defineE = 0",
   "enumC :: GLenum",
   "enumC = 42",
   "enumH :: GLenum",
   "enumH = 4"
 ]

check :: Show a => String -> a -> String -> IO ()
check name calculated expected = do
   let calculatedString = show calculated
   putStrLn ("-- " ++ name ++ "-----------------------------------------")
   putStrLn calculatedString
   if (calculatedString /= expected)
      then do
         putStrLn "-- EXPECTED -----------------------------------------"
         putStrLn expected
         error (name ++ " test failed")
      else return ()

test :: IO ()
test = do
   let parsed = parseSpec "test" (preprocess testSpecString)
   check "parsing" parsed testSpecString

   let expanded = expandSpec parsed
   check "expansion" expanded testExpandedString

   let evaluated = evaluate expanded
   check "evaluation" (fst evaluated) testEvaluatedString

   let simplified = simplify evaluated
   check "simplification" simplified testSimplifiedString

   let renamed = rename simplified
   check "renaming" renamed testRenamedString

   let code = codeGen renamed
   check "code generation" code testCodeString
