\section[IORef]{Mutable Variables}

\begin{code}
module IORef
	( IORef		      -- abstract, instance of: Eq
        , newIORef	      -- :: a -> IO (IORef a)
        , readIORef	      -- :: IORef a -> IO a
        , writeIORef	      -- :: IORef a -> a -> IO ()
	, modifyIORef	      -- :: IORef a -> (a -> a) -> IO ()
	, updateIORef	      -- deprecated, use modifyIORef

#if !defined(__PARALLEL_HASKELL__) && defined(__GLASGOW_HASKELL__)
	, mkWeakIORef           -- :: IORef a -> IO () -> IO (Weak (IORef a))
#endif
	) where

#ifdef __GLASGOW_HASKELL__
import PrelGHC		( mkWeak# )
import PrelArr		( STRef(..) )
import PrelIOBase
#if !defined(__PARALLEL_HASKELL__)
import PrelWeak
#endif
#endif
\end{code}

\begin{code}
#if defined __GLASGOW_HASKELL__ && !defined __PARALLEL_HASKELL__
mkWeakIORef :: IORef a -> IO () -> IO (Weak (IORef a))
mkWeakIORef r@(IORef (STRef r#)) f = IO $ \s ->
  case mkWeak# r# r f s of (# s1, w #) -> (# s1, Weak w #)
#endif

#if defined __HUGS__
data IORef a        -- mutable variables containing values of type a

primitive newIORef   "newRef" :: a -> IO (IORef a)
primitive readIORef  "getRef" :: IORef a -> IO a
primitive writeIORef "setRef" :: IORef a -> a -> IO ()
primitive eqIORef    "eqRef"  :: IORef a -> IORef a -> Bool

instance Eq (IORef a) where
    (==) = eqIORef

modifyIORef :: IORef a -> (a -> a) -> IO ()
modifyIORef ref f = writeIORef ref . f =<< readIORef ref

-- deprecated, use modifyIORef
updateIORef :: IORef a -> (a -> a) -> IO ()
updateIORef = modifyIORef
#endif
\end{code}
