{-# OPTIONS -cpp #-}

module DeviceSystemState (module DeviceSystemState, module Device) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	DeviceSystemState defines the device administration types and unwrapper 
--	functions.
--	********************************************************************************


import Device
import ReceiverHandle
import WindowHandle
import TimerHandle
import MenuHandle
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


data	DeviceSystemState IF_MVAR(,ps)
	= MenuSystemState	(MenuHandles		ps)
	| ReceiverSystemState IF_MVAR(ReceiverHandles,(ReceiverHandles ps))
	| TimerSystemState	(TimerHandles		ps)
	| WindowSystemState   IF_MVAR(WindowHandles,(WindowHandles ps))

toDevice :: DeviceSystemState IF_MVAR(,ps) -> Device
toDevice (ReceiverSystemState _) = ReceiverDevice
toDevice (WindowSystemState   _) = WindowDevice
toDevice (TimerSystemState   _)  = TimerDevice
toDevice (MenuSystemState     _) = MenuDevice


receiverSystemStateGetReceiverHandles :: DeviceSystemState IF_MVAR(,ps) -> ReceiverHandles IF_MVAR(,ps)
receiverSystemStateGetReceiverHandles (ReceiverSystemState rsHs) = rsHs 

windowSystemStateGetWindowHandles :: DeviceSystemState IF_MVAR(,ps) -> WindowHandles IF_MVAR(,ps)
windowSystemStateGetWindowHandles (WindowSystemState wsHs) = wsHs

timerSystemStateGetTimerHandles :: DeviceSystemState IF_MVAR(,ps) -> TimerHandles IF_MVAR(,ps)
timerSystemStateGetTimerHandles (TimerSystemState tsHs) = tsHs

menuSystemStateGetMenuHandles :: DeviceSystemState ps -> MenuHandles ps
menuSystemStateGetMenuHandles (MenuSystemState msHs) = msHs

