{-# OPTIONS -cpp #-}

module StdControlDef ( module StdControlDef
                     , module StdIOCommon
                     , module StdGUI
                     ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--
--	StdControlDef contains the types to define the standard set of controls.
--	********************************************************************************


import StdGUI
import StdIOCommon
import StdPicture(Look)
import StdPictureDef
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


data	ButtonControl IF_MVAR(,ls ps)
 =	ButtonControl String                      				[ControlAttribute IF_MVAR(,ls ps)]

data	CheckControl  IF_MVAR(,ls ps)
 = 	CheckControl  [CheckControlItem ps (ls,ps)]  RowsOrColumns       	[ControlAttribute IF_MVAR(,ls ps)]

data	CompoundControl c IF_MVAR(,ls ps)
 = 	CompoundControl IF_MVAR(c,(c ls ps))                                	[ControlAttribute IF_MVAR(,ls ps)]

data	CustomButtonControl IF_MVAR(,ls ps)
 = 	CustomButtonControl Size Look                                        	[ControlAttribute IF_MVAR(,ls ps)]

data	CustomControl       IF_MVAR(,ls ps)
 = 	CustomControl       Size Look                                        	[ControlAttribute IF_MVAR(,ls ps)]

data	EditControl   IF_MVAR(,ls ps)
 =	EditControl   String ControlWidth NrLines 				[ControlAttribute IF_MVAR(,ls ps)]

data	LayoutControl     c IF_MVAR(,ls ps)
 = 	LayoutControl     IF_MVAR(c,(c ls ps))                                  [ControlAttribute IF_MVAR(,ls ps)]

data	PopUpControl 	IF_MVAR(,ls ps)
 = 	PopUpControl   	[PopUpControlItem ps (ls,ps)] Index 			[ControlAttribute IF_MVAR(,ls ps)]

data	RadioControl   	IF_MVAR(,ls ps)
 = 	RadioControl    [RadioControlItem ps (ls,ps)] RowsOrColumns Index 	[ControlAttribute IF_MVAR(,ls ps)]

data	SliderControl  	IF_MVAR(,ls ps)
 = 	SliderControl   Direction ControlWidth SliderState  (SliderAction  IF_MVAR(,ls ps)) [ControlAttribute IF_MVAR(,ls ps)]

data	TextControl   IF_MVAR(,ls ps)
 =	TextControl   String                      				[ControlAttribute IF_MVAR(,ls ps)]


type	CheckControlItem ps st = (String, Maybe ControlWidth, MarkState, st -> GUI ps st)
type	PopUpControlItem ps st = (String,                                st -> GUI ps st)
type	RadioControlItem ps st = (String, Maybe ControlWidth,            st -> GUI ps st)


type	NrLines
 =	Int
data	RowsOrColumns
	= Rows       Int
	| Columns    Int
data	ControlWidth                                -- The width of the control:
 =	PixelWidth   Int                            -- the exact number of pixels
 |	TextWidth    String                         -- the exact string width in dialog font
 |	ContentWidth String                         -- width of the control as if string is its content

data	ControlAttribute IF_MVAR(,ls ps)            -- Default:
 -- General control attributes:
 =	ControlActivate     (GUIFun IF_MVAR((),ls ps))  -- id
 |	ControlDeactivate   (GUIFun IF_MVAR((),ls ps))  -- id
 |	ControlFunction (GUIFun IF_MVAR((),ls ps))  -- (\st->return st)
 |	ControlHide                                 -- initially visible
 |	ControlId       Id                          -- no id
 |	ControlKeyboard KeyboardStateFilter SelectState IF_MVAR(KeyboardFunction,(KeyboardFunction ls ps))
	                                            -- no keyboard input/overruled
 |	ControlMinimumSize  Size                    -- zero
 |	ControlModsFunction (ModifiersFunction IF_MVAR(,ls ps))
                                                    -- ControlFunction
 |	ControlMouse        MouseStateFilter    SelectState (MouseFunction IF_MVAR(,ls ps))
		                                    -- no mouse input/overruled
 |	ControlPen	    [PenAttribute]	    -- default pen attributes
 |	ControlPos          ItemPos                 -- (RightTo previous,zero)
 |	ControlResize       ControlResizeFunction   -- no resize
 |	ControlSelectState  SelectState             -- control Able
 |	ControlTip          String                  -- no tip
 |	ControlWidth        ControlWidth            -- system derived
 --	For CompoundControls only:
 |	ControlHMargin      Int Int                 -- system dependent
 |	ControlHScroll      ScrollFunction          -- no horizontal scrolling
 |	ControlItemSpace    Int Int                 -- system dependent
 |	ControlLook         Bool Look               -- control is transparant
 |	ControlOrigin       Point2                  -- Left top of ViewDomain
 |	ControlOuterSize    Size		    -- enclose elements
 |	ControlViewDomain   ViewDomain              -- {zero,max range}
 |	ControlViewSize     Size                    -- enclose elements
 |	ControlVMargin      Int Int                 -- system dependent
 |	ControlVScroll      ScrollFunction          -- no vertical   scrolling




type ControlResizeFunction =
	Size ->                                     		-- current control outer size
	Size ->                                     		-- old     parent  view  size
	Size ->                                     		-- new     parent  view  size
	Size                                        		-- new     control outer size



type ControlType = String