{-# OPTIONS -cpp #-}

module StdWindowAttribute ( isValidWindowAttribute, isValidDialogAttribute
                          , isWindowActivate,   getWindowActivateFun
                          , isWindowCancel,	getWindowCancelAtt
                          , isWindowClose,      getWindowCloseFun
                          , isWindowCursor,	getWindowCursorAtt
                          , isWindowDeactivate, getWindowDeactivateFun
                          , isWindowHMargin,	getWindowHMarginAtt
                          , isWindowHScroll,	getWindowHScrollFun
                          , isWindowId,         getWindowIdAtt
                          , isWindowIndex,	getWindowIndexAtt
                          , isWindowInit,       getWindowInitFun
                          , isWindowInitActive, getWindowInitActiveAtt
                          , isWindowItemSpace,  getWindowItemSpaceAtt
                          , isWindowKeyboard,	getWindowKeyboardAtt
                          , isWindowLook,	getWindowLookAtt
                          , isWindowMouse,	getWindowMouseAtt
                          , isWindowOk,		getWindowOkAtt
                          , isWindowOrigin,	getWindowOriginAtt
                          , isWindowOuterSize,	getWindowOuterSizeAtt
                          , isWindowPen,	getWindowPenAtt
                          , isWindowPos,	getWindowPosAtt
                          , isWindowSelectState,getWindowSelectStateAtt
                          , isWindowViewDomain, getWindowViewDomainAtt
                          , isWindowViewSize,   getWindowViewSizeAtt
                          , isWindowVMargin,	getWindowVMarginAtt
                          , isWindowVScroll,	getWindowVScrollFun
                          , module StdWindowDef
                          ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	StdWindowAttribute specifies which WindowAttributes are valid for Windows.
--	********************************************************************************


import StdWindowDef

#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


isValidWindowAttribute :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isValidWindowAttribute att = isAllWindowsAttribute att || isWindowOnlyAttribute att

isValidDialogAttribute :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isValidDialogAttribute att = isAllWindowsAttribute att || isDialogOnlyAttribute att

isAllWindowsAttribute :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isAllWindowsAttribute (WindowActivate   _) = True
isAllWindowsAttribute (WindowClose      _) = True
isAllWindowsAttribute (WindowDeactivate _) = True
isAllWindowsAttribute (WindowHMargin  _ _) = True
isAllWindowsAttribute (WindowId         _) = True
isAllWindowsAttribute (WindowIndex	_) = True
isAllWindowsAttribute (WindowInit       _) = True
isAllWindowsAttribute (WindowInitActive _) = True
isAllWindowsAttribute (WindowItemSpace _ _)= True
isAllWindowsAttribute (WindowOuterSize	_) = True
isAllWindowsAttribute (WindowPos	_) = True
isAllWindowsAttribute (WindowViewSize   _) = True
isAllWindowsAttribute (WindowVMargin  _ _) = True
isAllWindowsAttribute _                    = False

isWindowOnlyAttribute :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowOnlyAttribute (WindowCursor	  _) = True
isWindowOnlyAttribute (WindowHScroll	  _) = True
isWindowOnlyAttribute (WindowKeyboard _ _ _) = True
isWindowOnlyAttribute (WindowLook       _ _) = True
isWindowOnlyAttribute (WindowMouse    _ _ _) = True
isWindowOnlyAttribute (WindowOrigin	  _) = True
isWindowOnlyAttribute (WindowPen          _) = True
isWindowOnlyAttribute (WindowSelectState  _) = True
isWindowOnlyAttribute (WindowViewDomain	  _) = True
isWindowOnlyAttribute (WindowVScroll	  _) = True
isWindowOnlyAttribute _			     = False

isDialogOnlyAttribute :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isDialogOnlyAttribute (WindowCancel	  _) = True
isDialogOnlyAttribute (WindowOk		  _) = True
isDialogOnlyAttribute _			     = False

isWindowActivate :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowActivate (WindowActivate _) = True
isWindowActivate _                  = False

isWindowCancel	 :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowCancel	(WindowCancel _) = True
isWindowCancel	_		 = False

isWindowClose    :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowClose    (WindowClose _)    = True
isWindowClose    _                  = False

isWindowCursor	:: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowCursor	(WindowCursor _) = True
isWindowCursor	_		 = False

isWindowDeactivate :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowDeactivate (WindowDeactivate _) = True
isWindowDeactivate _                    = False

isWindowHMargin	:: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowHMargin	(WindowHMargin _ _) = True
isWindowHMargin	_		    = False

isWindowHScroll	:: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowHScroll	(WindowHScroll _) = True
isWindowHScroll	_		  = False

isWindowId :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowId (WindowId _) = True
isWindowId _            = False

isWindowIndex :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowIndex (WindowIndex _) = True
isWindowIndex _		      = False

isWindowInit :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowInit (WindowInit _) = True
isWindowInit _              = False

isWindowInitActive :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowInitActive (WindowInitActive _)	= True
isWindowInitActive _			= False

isWindowItemSpace :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowItemSpace (WindowItemSpace _ _)	= True
isWindowItemSpace _			= False

isWindowKeyboard :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowKeyboard (WindowKeyboard _ _ _)	= True
isWindowKeyboard _						= False

isWindowLook :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowLook (WindowLook _ _) = True
isWindowLook _		      = False

isWindowMouse :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowMouse (WindowMouse _ _ _) = True
isWindowMouse _			  = False

isWindowOk :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowOk (WindowOk _)	= True
isWindowOk _		= False

isWindowOrigin :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowOrigin (WindowOrigin _)	= True
isWindowOrigin _		= False

isWindowOuterSize :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowOuterSize (WindowOuterSize _) = True
isWindowOuterSize _		      = False

isWindowPen :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowPen (WindowPen _) = True
isWindowPen _		  = False

isWindowPos :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowPos (WindowPos _) = True
isWindowPos _		  = False

isWindowSelectState :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowSelectState (WindowSelectState _) = True
isWindowSelectState _			  = False

isWindowViewDomain :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowViewDomain (WindowViewDomain _)	= True
isWindowViewDomain _			= False

isWindowViewSize :: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowViewSize (WindowViewSize _) = True
isWindowViewSize _                  = False

isWindowVMargin	:: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowVMargin	(WindowVMargin _ _) = True
isWindowVMargin	_		    = False

isWindowVScroll	:: WindowAttribute IF_MVAR(,ls ps) -> Bool
isWindowVScroll	(WindowVScroll _) = True
isWindowVScroll	_		  = False

getWindowActivateFun :: WindowAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getWindowActivateFun (WindowActivate f) = f

getWindowCancelAtt :: WindowAttribute IF_MVAR(,ls ps) -> Id
getWindowCancelAtt (WindowCancel id) = id

getWindowCloseFun :: WindowAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getWindowCloseFun (WindowClose f) = f

getWindowCursorAtt :: WindowAttribute IF_MVAR(,ls ps) -> CursorShape
getWindowCursorAtt (WindowCursor cShape) = cShape

getWindowDeactivateFun :: WindowAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getWindowDeactivateFun (WindowDeactivate f) = f

getWindowHMarginAtt :: WindowAttribute IF_MVAR(,ls ps) -> (Int,Int)
getWindowHMarginAtt (WindowHMargin left right) = (left,right)

getWindowHScrollFun :: WindowAttribute IF_MVAR(,ls ps) -> ScrollFunction
getWindowHScrollFun (WindowHScroll f) = f

getWindowIdAtt :: WindowAttribute IF_MVAR(,ls ps) -> Id
getWindowIdAtt (WindowId id) = id

getWindowIndexAtt :: WindowAttribute IF_MVAR(,ls ps) -> Int
getWindowIndexAtt (WindowIndex index) = index

getWindowInitFun :: WindowAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getWindowInitFun (WindowInit init) = init

getWindowInitActiveAtt :: WindowAttribute IF_MVAR(,ls ps) -> Id
getWindowInitActiveAtt (WindowInitActive id) = id

getWindowItemSpaceAtt :: WindowAttribute IF_MVAR(,ls ps) -> (Int,Int)
getWindowItemSpaceAtt (WindowItemSpace hspace vspace) = (hspace,vspace)

getWindowKeyboardAtt :: WindowAttribute IF_MVAR(,ls ps) -> (KeyboardStateFilter,SelectState,KeyboardFunction IF_MVAR(,ls ps))
getWindowKeyboardAtt (WindowKeyboard filter select keysF) = (filter,select,keysF)

getWindowLookAtt :: WindowAttribute IF_MVAR(,ls ps) -> (Bool,Look)
getWindowLookAtt (WindowLook systemLook f) = (systemLook,f)

getWindowMouseAtt :: WindowAttribute IF_MVAR(,ls ps)  -> (MouseStateFilter,SelectState,MouseFunction IF_MVAR(,ls ps))
getWindowMouseAtt (WindowMouse filter select mouseF) = (filter,select,mouseF)

getWindowOkAtt :: WindowAttribute IF_MVAR(,ls ps) -> Id
getWindowOkAtt (WindowOk id) = id

getWindowOriginAtt :: WindowAttribute IF_MVAR(,ls ps) -> Point2
getWindowOriginAtt (WindowOrigin origin) = origin

getWindowOuterSizeAtt :: WindowAttribute IF_MVAR(,ls ps) -> Size
getWindowOuterSizeAtt (WindowOuterSize size) = size

getWindowPenAtt :: WindowAttribute IF_MVAR(,ls ps) -> [PenAttribute]
getWindowPenAtt (WindowPen pen) = pen

getWindowPosAtt :: WindowAttribute IF_MVAR(,ls ps) -> ItemPos
getWindowPosAtt (WindowPos pos) = pos

getWindowSelectStateAtt :: WindowAttribute IF_MVAR(,ls ps) -> SelectState
getWindowSelectStateAtt (WindowSelectState select) = select

getWindowViewDomainAtt :: WindowAttribute IF_MVAR(,ls ps) -> ViewDomain
getWindowViewDomainAtt (WindowViewDomain d) = d

getWindowViewSizeAtt :: WindowAttribute IF_MVAR(,ls ps) -> Size
getWindowViewSizeAtt (WindowViewSize size) = size

getWindowVMarginAtt :: WindowAttribute IF_MVAR(,ls ps) -> (Int,Int)
getWindowVMarginAtt (WindowVMargin top bottom) = (top,bottom)

getWindowVScrollFun :: WindowAttribute IF_MVAR(,ls ps) -> ScrollFunction
getWindowVScrollFun (WindowVScroll f) = f
