module TimerDevice(timerFunctions) where


--	Clean Object I/O library, version 1.2

import	CommonDef
import	IOState
import	ReceiverHandle
import	ReceiverTable
import  StdTimerDef(NrOfIntervals)
import	TimerAccess
import	TimerDefAccess
import	TimerEvent
import	TimerHandle
import	TimerTable
import  Id
import  IOExts(fixIO)


timerDeviceFatalError :: String -> String -> x
timerDeviceFatalError function error = dumpFatalError function "TimerDevice" error

timerFunctions :: DeviceFunctions ps
timerFunctions =
	DeviceFunctions 
	  {	dDevice	= TimerDevice
	  ,	dShow	= return
	  ,	dHide	= return
	  ,	dEvent	= timerEvent
	  ,	dDoIO	= timerIO
	  ,	dOpen	= timerOpen
	  ,	dClose	= timerClose
	  }

timerOpen :: ps -> GUI ps ps
timerOpen ps = do
	hasTimer <- accIOEnv (ioStHasDevice TimerDevice)
	(if hasTimer then return ps
	 else do
		appIOEnv (ioStSetDevice (TimerSystemState (TimerHandles	{tTimers=[]})))
		appIOEnv (ioStSetDeviceFunctions timerFunctions)
		return ps)
		

timerClose :: ps -> GUI ps ps
timerClose ps = do
	(found,timers) <- accIOEnv (ioStGetDevice TimerDevice)
	(if not found then return ps
	 else do
		rt <- ioStGetReceiverTable
		tt <- ioStGetTimerTable
		it <- ioStGetIdTable
		pid <- accIOEnv ioStGetIOId
		let tHs	= timerSystemStateGetTimerHandles timers
		let (tt1,rt1,it1) = foldr (closeTimerIds pid) (tt,rt,it) (tTimers tHs)
		ioStSetReceiverTable rt1
		ioStSetTimerTable tt1
		ioStSetIdTable it1
		appIOEnv (ioStRemoveDevice TimerDevice)
		appIOEnv (ioStRemoveDeviceFunctions TimerDevice)
		return ps)
	where
		closeTimerIds :: SystemId -> TimerStateHandle ps -> (TimerTable,ReceiverTable,IdTable) -> (TimerTable,ReceiverTable,IdTable)
		closeTimerIds pid (TimerStateHandle (TimerLSHandle {tHandle=(TimerHandle {tId=tId,tItems=tItems})})) tables =
			let (tt,rt,it)	= unbindTimerElementIds pid tItems tables
			in (snd (removeTimerFromTimerTable teLoc tt),rt,snd (removeIdFromIdTable tId it))
			where
				teLoc = TimerLoc{tlIOId=pid,tlDevice=TimerDevice,tlParentId=tId,tlTimerId=tId}


timerIO	:: DeviceEvent -> ps -> GUI ps ps
timerIO deviceEvent ps = do
	hasDevice <- accIOEnv (ioStHasDevice TimerDevice)
	(if (not hasDevice)
	 then timerDeviceFatalError "timerFunctions.dDoIO" "could not retrieve TimerSystemState from IOSt"
	 else timerIO deviceEvent ps)
	where
		timerIO	:: DeviceEvent -> ps -> GUI ps ps
		timerIO (TimerDeviceEvent (TimerEvent {teLoc=TimerLoc{tlParentId=tlParentId,tlTimerId=tlTimerId},teNrInterval=teNrInterval})) ps = do
			(_,timer) <- accIOEnv (ioStGetDevice TimerDevice)
			let timers = timerSystemStateGetTimerHandles timer		  
			toGUI (letOneTimerDoIO tlParentId tlTimerId teNrInterval timers ps)
			where
				letOneTimerDoIO :: Id -> Id -> NrOfIntervals -> TimerHandles ps -> ps -> IOSt ps -> IO (ps, IOSt ps)
				letOneTimerDoIO parent timer nrOfIntervals timers ps ioState = f tsH ioState
					where
						(_,tsH,tsHs) = remove (identifyTimerStateHandle parent) (timerDeviceFatalError "timerIO (TimerEvent _)" "timer could not be found") (tTimers timers)

						f (TimerStateHandle tlsH@(TimerLSHandle {tState=ls,tHandle=tH})) ioState = do
							r <- fixIO (\st -> fromGUI (tFun tH nrOfIntervals (ls,ps))
								(ioStSetDevice (TimerSystemState timers{tTimers=tsHs++[TimerStateHandle tlsH{tState=fst (fst st)}]}) ioState))
							let ((_,ps1), ioState) = r
							return (ps1,ioState)
						
{-
	timerIO deviceEvent=:(ReceiverEvent (QASyncMessage event)) pState
		= (deviceEvent,timerQASync event pState)
	where
		timerQASync :: !QASyncMessage !(PSt .l) -> PSt .l
		timerQASync msg=:{qasmRecLoc={rlReceiverId=rid}} pState
			# (_,timer,ioState)	= IOStGetDevice TimerDevice pState.io
			  timers			= TimerSystemStateGetTimerHandles timer
			  tsHs				= handleASyncReceiver rid msg timers.tTimers
			  timers			= {timers & tTimers=tsHs}
			# ioState			= IOStSetDevice (TimerSystemState timers) ioState
			# pState			= {pState & io=ioState}
			= pState
		where
			handleASyncReceiver :: !Id !QASyncMessage ![TimerStateHandle .pst] -> [TimerStateHandle .pst]
			handleASyncReceiver rid msg [TimerLSHandle tlsH=:{tHandle=tH=:{tItems}}:tsHs]
				# (done,tItems)	= qMessage rid msg tItems
				  tsH			= TimerLSHandle {tlsH & tHandle={tH & tItems=tItems}}
				| done			= [tsH:tsHs]
				| otherwise		= [tsH:handleASyncReceiver rid msg tsHs]
			where
				qMessage :: !Id !QASyncMessage ![TimerElementHandle .ls .pst] -> (!Bool,![TimerElementHandle .ls .pst])
				qMessage rid msg [itemH:itemHs]
					# (done,itemH)	= qMessage` rid msg itemH
					| done
						= (done,[itemH:itemHs])
					| otherwise
						# (done,itemHs)	= qMessage rid msg itemHs
						= (done,[itemH:itemHs])
				where
					qMessage` :: !Id !QASyncMessage !(TimerElementHandle .ls .pst) -> (!Bool,!TimerElementHandle .ls .pst)
					qMessage` rid {qasmMsg} (TimerReceiverHandle trH=:{tReceiverHandle=rH})
						| receiverIdentified rid rH	= (True, TimerReceiverHandle {trH & tReceiverHandle=receiverAddASyncMessage rid qasmMsg rH})
						| otherwise					= (False,TimerReceiverHandle trH)
					qMessage` rid msg (TimerListLSHandle itemHs)
						# (done,itemHs) = qMessage rid msg itemHs
						= (done,TimerListLSHandle itemHs)
					qMessage` rid msg (TimerElimLSHandle itemHs)
						# (done,itemHs) = qMessage rid msg itemHs
						= (done,TimerElimLSHandle itemHs)
					qMessage` rid msg (TimerIntroLSHandle tInH=:{tIntroItems})
						# (done,itemHs) = qMessage rid msg tIntroItems
						= (done,TimerIntroLSHandle {tInH & tIntroItems=itemHs})
					qMessage` rid msg (TimerExtendLSHandle tExH=:{tExtendItems})
						# (done,itemHs) = qMessage rid msg tExtendItems
						= (done,TimerExtendLSHandle {tExH & tExtendItems=itemHs})
					qMessage` rid msg (TimerChangeLSHandle tChH=:{tChangeItems})
						# (done,itemHs) = qMessage rid msg tChangeItems
						= (done,TimerChangeLSHandle {tChH & tChangeItems=itemHs})
				qMessage _ _ _
					= (False,[])
			handleASyncReceiver _ _ _
				= []
-}
		timerIO (ReceiverEvent msg@(ASyncMessage {})) ps = timerASync msg ps
			where
				timerASync :: MsgEvent -> ps -> GUI ps ps
				timerASync (ASyncMessage{asmRecLoc=RecLoc{rlReceiverId=rid}}) ps = do
					(_,timers) <- accIOEnv (ioStGetDevice TimerDevice)
					let tHs = timerSystemStateGetTimerHandles timers
					toGUI (aSyncTimerStateHandles (\tsHs st ioState -> ioStSetDevice (TimerSystemState tHs{tTimers=tsHs}) ioState) rid (tTimers tHs) ps)
					where			
						aSyncTimerStateHandles :: ([TimerStateHandle ps] -> ps -> IOSt ps -> IOSt ps) ->
									  Id -> [TimerStateHandle ps] -> ps -> IOSt ps -> IO (ps,IOSt ps)
						aSyncTimerStateHandles build rid (tsH@(TimerStateHandle (TimerLSHandle {tState=ls,tHandle=tH@(TimerHandle {tItems=itemHs})})):tsHs) ps ioState = do
							r <- aSyncTimerElementHandles (\itemHs st ioState -> build (TimerStateHandle (TimerLSHandle {tState=fst st,tHandle=tH{tItems=itemHs}}):tsHs) (snd st) ioState) rid itemHs (ls,ps) ioState
							let (done,(ls,ps),ioState) = r				  
							(if done then return (ps, ioState)
							 else aSyncTimerStateHandles (\tsHs st ioState -> build (tsH:tsHs) st ioState) rid tsHs ps ioState)
							where
								aSyncTimerElementHandles :: ([TimerElementHandle ls ps] -> (ls,ps) -> IOSt ps -> IOSt ps) ->
											    Id -> [TimerElementHandle ls ps] -> (ls,ps) -> IOSt ps -> IO (Bool,(ls,ps),IOSt ps)
								aSyncTimerElementHandles build rid (itemH:itemHs) ls_ps ioState = do
									r <- aSyncTimerElementHandle (\itemH st ioState -> build (itemH:itemHs) st ioState) rid itemH ls_ps ioState
									let (done,ls_ps,ioState) = r
									(if done then return r
									 else aSyncTimerElementHandles (\itemHs st ioState -> build (itemH:itemHs) st ioState) rid itemHs ls_ps ioState)
									where
										aSyncTimerElementHandle :: (TimerElementHandle ls ps -> (ls,ps) -> IOSt ps -> IOSt ps) ->
													   Id -> TimerElementHandle ls ps -> (ls,ps) -> IOSt ps -> IO (Bool,(ls,ps),IOSt ps)
										aSyncTimerElementHandle build rid (TimerListLSHandle itemHs) ls_ps ioState =
											aSyncTimerElementHandles (\itemHs st ioState -> build (TimerListLSHandle itemHs) st ioState) rid itemHs ls_ps ioState

										aSyncTimerElementHandle build rid (TimerElimLSHandle itemHs) ls_ps ioState =
											aSyncTimerElementHandles (\itemHs st ioState -> build (TimerElimLSHandle itemHs) st ioState) rid itemHs ls_ps ioState

										aSyncTimerElementHandle build rid (TimerIntroLSHandle introLS itemHs) (ls,ps) ioState = do
											r <- aSyncTimerElementHandles (\itemHs st ioState -> build (TimerIntroLSHandle (fst st) itemHs) (ls,snd st) ioState) rid itemHs (introLS,ps) ioState
											let (done,(_,ps),ioState) = r
											return (done,(ls,ps),ioState)

										aSyncTimerElementHandle build rid (TimerExtendLSHandle exLS itemHs) (ls,ps) ioState = do
											r <- aSyncTimerElementHandles (\itemHs st ioState -> build (TimerExtendLSHandle (fst (fst st)) itemHs) (snd (fst st), snd st) ioState) rid itemHs ((exLS,ls),ps) ioState
											let (done,((_,ls),ps),ioState) = r
											return (done,(ls,ps),ioState)

										aSyncTimerElementHandle build rid (TimerChangeLSHandle chLS itemHs) (ls,ps) ioState = do
											r <- aSyncTimerElementHandles (\itemHs st ioState -> build (TimerChangeLSHandle (fst st) itemHs) (ls,snd st) ioState) rid itemHs (chLS,ps) ioState
											let (done,(_,ps),ioState) = r
											return (done,(ls,ps),ioState)

										aSyncTimerElementHandle build rid itemH@(TimerReceiverHandle rH@(ReceiverHandle {rFun=rFun,rASMQ=channel}) atts) ls_ps ioState
											| receiverIdentified rid rH	= do
												msg <- readChan channel
												r <- fixIO (\st -> fromGUI (rFun msg ls_ps) (build (TimerReceiverHandle rH atts) (snd (fst st)) ioState))
												let ((_, ls_ps), ioState) = r
												return (True, ls_ps, ioState)
											| otherwise			=
												return (False,ls_ps, ioState)

								aSyncTimerElementHandles _ _ _ ps ioState = return (False,ps,ioState)
						aSyncTimerStateHandles _ _ _ ps ioState = return (ps,ioState)
{-
	timerIO (ReceiverEvent (SyncMessage event)) pState
		# (event,pState)	= timerSync event pState
		= (ReceiverEvent (SyncMessage event),pState)
	where
		timerSync :: !SyncMessage !(PSt .l) -> (!SyncMessage,!PSt .l)
		timerSync msg=:{smRecLoc={rlReceiverId=rid}} pState
			= (msg1,pState2)
		where
			(_,timers,ioState)	= IOStGetDevice TimerDevice pState.io
			tHs					= TimerSystemStateGetTimerHandles timers
			(msg1,tsHs,pState2)	= syncTimerStateHandles rid msg tHs.tTimers pState1
			tHs1				= {tHs & tTimers=tsHs}
			ioState1			= IOStSetDevice (TimerSystemState tHs1) ioState
			pState1				= {pState & io=ioState1}
			
			syncTimerStateHandles :: !Id !SyncMessage ![TimerStateHandle .pst] .pst
									 -> (!SyncMessage,![TimerStateHandle .pst],.pst)
			syncTimerStateHandles rid msg [TimerLSHandle {tState=ls,tHandle=tH=:{tItems}}:tsHs] ps
				# (done,msg,tItems,(ls,ps))	= syncTimerElementHandles rid msg tItems (ls,ps)
				  tsH						= TimerLSHandle {tState=ls,tHandle={tH & tItems=tItems}}
				| done
					= (msg,[tsH:tsHs],ps)
				| otherwise
					# (msg,tsHs,ps)			= syncTimerStateHandles rid msg tsHs ps
					= (msg,[tsH:tsHs],ps)
			where
				syncTimerElementHandles :: !Id !SyncMessage ![TimerElementHandle .ls .pst] (.ls,.pst)
									 -> (!Bool,!SyncMessage,![TimerElementHandle .ls .pst],(.ls,.pst))
				syncTimerElementHandles rid msg [itemH:itemHs] (ls,ps)
					# (done,msg,itemH,(ls,ps))	= syncTimerElementHandle rid msg itemH (ls,ps)
					| done
						= (done,msg,[itemH:itemHs],(ls,ps))
					| otherwise
						# (done,msg,itemHs,(ls,ps))	= syncTimerElementHandles rid msg itemHs (ls,ps)
						= (done,msg,[itemH:itemHs],(ls,ps))
				where
					syncTimerElementHandle :: !Id !SyncMessage !(TimerElementHandle .ls .pst) (.ls,.pst)
										-> (!Bool,!SyncMessage, !TimerElementHandle .ls .pst, (.ls,.pst))
					syncTimerElementHandle rid msg (TimerReceiverHandle trH=:{tReceiverHandle=rH}) (ls,ps)
						| not (receiverIdentified rid rH)
							= (False,msg,TimerReceiverHandle trH,(ls,ps))
						| otherwise
							# (resp,rH,(ls,ps))	= receiverHandleSyncMessage msg rH (ls,ps)
							= (True, {msg & smResp=resp},TimerReceiverHandle {trH & tReceiverHandle=rH},(ls,ps))
					syncTimerElementHandle rid msg (TimerListLSHandle itemHs) (ls,ps)
						# (done,msg,itemHs,(ls,ps))	= syncTimerElementHandles rid msg itemHs (ls,ps)
						= (done,msg,TimerListLSHandle itemHs,(ls,ps))
					syncTimerElementHandle rid msg (TimerElimLSHandle itemHs) (ls,ps)
						# (done,msg,itemHs,(ls,ps))	= syncTimerElementHandles rid msg itemHs (ls,ps)
						= (done,msg,TimerElimLSHandle itemHs,(ls,ps))
					syncTimerElementHandle rid msg (TimerIntroLSHandle {tIntroLS=inLS,tIntroItems}) (ls,ps)
						# (done,msg,itemHs,(inLS,ps))	= syncTimerElementHandles rid msg tIntroItems (inLS,ps)
						= (done,msg,TimerIntroLSHandle {tIntroLS=inLS,tIntroItems=itemHs},(ls,ps))
					syncTimerElementHandle rid msg (TimerExtendLSHandle {tExtendLS,tExtendItems}) (ls,ps)
						# (done,msg,itemHs,((exLS,ls),ps))	= syncTimerElementHandles rid msg tExtendItems ((tExtendLS,ls),ps)
						= (done,msg,TimerExtendLSHandle {tExtendLS=exLS,tExtendItems=itemHs},(ls,ps))
					syncTimerElementHandle rid msg (TimerChangeLSHandle {tChangeLS,tChangeItems}) (ls,ps)
						# (done,msg,itemHs,(chLS,ps))	= syncTimerElementHandles rid msg tChangeItems (tChangeLS,ps)
						= (done,msg,TimerChangeLSHandle {tChangeLS=chLS,tChangeItems=itemHs},(ls,ps))
				syncTimerElementHandles _ msg _ (ls,ps)
					= (False,msg,[],(ls,ps))
			syncTimerStateHandles _ msg _ ps
				= (msg,[],ps)
	timerIO _ _
		= timerdeviceFatalError "timerIO" "device event passed timer event filter without handling"
-}